[CCCL] 瘦身 + 补全: 移除 cudax/python/libcudacxx-tests 冗余文件, 新增 c2h 测试助手 + cmake 构建系统 + 8 个 CUDA thrust examples

变更摘要:
- 删除: cudax/ (783 files, 7.2M) — 实验性组件,竞赛不需要
- 删除: python/ (226 files, 2.0M) — Python 绑定,竞赛不需要
- 删除: libcudacxx/{test,benchmarks,codegen,cmake,share} (4432 files, 31M)
  保留: libcudacxx/include/ (1463 headers, cuda::std 编译依赖)
- 新增: c2h/ (27 files) — CUB Catch2 测试辅助头文件,编译 243 个测试必需
- 新增: cmake/ (29 files) — CCCL 原生 CMake 构建系统
- 新增: thrust/examples/cuda/ (7 files) + cpp_integration/ (1 file)
  async_reduce, custom_temporary_allocation, explicit_cuda_stream,
  global_device_vector, range_view, unwrap_pointer, wrap_pointer, device

结果: cccl_upstream 从 74M→35M (瘦身 53%), 核心内容 100% 保留:
  27/27 tuning headers, 78 benchmarks, 243 tests,
  60 thrust examples, 18 CUB examples, 全部编译头文件
This commit is contained in:
muh-bot
2026-08-03 12:39:26 +00:00
parent a2a5dd8f00
commit 24ef6a91b5
5439 changed files with 0 additions and 719516 deletions

View File

@@ -1,209 +0,0 @@
//===----------------------------------------------------------------------===//
//
// Part of CUDA Experimental in CUDA C++ Core Libraries,
// under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES.
//
//===----------------------------------------------------------------------===//
#include <cuda/devices>
#include <cuda/hierarchy>
#include <cuda/launch>
#include <cuda/std/cstddef>
#include <cuda/std/type_traits>
#include <cuda/std/utility>
#include <cuda/stream>
#include <cuda/experimental/group.cuh>
#include "group_testing.cuh"
namespace
{
struct AlwaysTruePredFn
{
template <class MappingResult>
__device__ bool operator()(MappingResult mapping_result)
{
return true;
}
};
struct AlwaysFalsePredFn
{
template <class MappingResult>
__device__ bool operator()(MappingResult mapping_result) noexcept
{
return false;
}
};
struct IsEvenPredFn
{
template <class MappingResult>
__device__ bool operator()(MappingResult mapping_result)
{
return mapping_result.unit_rank() % 2 == 0;
}
};
template <class Config>
__device__ void test_binary_partition(Config config)
{
// Always true predicate.
{
using Pred = AlwaysTruePredFn;
using Mapping = cudax::binary_partition<Pred>;
// Test constructor from pred_fn.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, Pred>);
cudax::binary_partition mapping{Pred{}};
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
!noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping{Pred{}};
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
static_assert(Result::static_group_count() == 2);
REQUIRE(result.group_count() == 2);
REQUIRE(result.group_rank() == 1);
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
REQUIRE(result.unit_count() == cuda::gpu_thread.count(cuda::warp));
REQUIRE(result.unit_rank() == cuda::gpu_thread.rank(cuda::warp));
REQUIRE(result.lane_mask() == cuda::device::lane_mask::all());
REQUIRE(result.is_valid());
static_assert(Result::is_always_exhaustive());
static_assert(!Result::is_always_contiguous());
}
}
// Always false predicate.
{
using Pred = AlwaysFalsePredFn;
using Mapping = cudax::binary_partition<Pred>;
// Test constructor from pred_fn.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, Pred>);
cudax::binary_partition mapping{Pred{}};
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping{Pred{}};
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
static_assert(Result::static_group_count() == 2);
REQUIRE(result.group_count() == 2);
REQUIRE(result.group_rank() == 0);
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
REQUIRE(result.unit_count() == cuda::gpu_thread.count(cuda::warp));
REQUIRE(result.unit_rank() == cuda::gpu_thread.rank(cuda::warp));
REQUIRE(result.lane_mask() == cuda::device::lane_mask::all());
REQUIRE(result.is_valid());
static_assert(Result::is_always_exhaustive());
static_assert(!Result::is_always_contiguous());
}
}
// True for even ranks predicate.
{
using Pred = IsEvenPredFn;
using Mapping = cudax::binary_partition<Pred>;
// Test constructor from pred_fn.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, Pred>);
cudax::binary_partition mapping{Pred{}};
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
!noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping{Pred{}};
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
static_assert(Result::static_group_count() == 2);
REQUIRE(result.group_count() == 2);
REQUIRE(result.group_rank() == (cuda::gpu_thread.rank(cuda::warp) % 2 == 0));
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
REQUIRE(result.unit_count() == cuda::gpu_thread.count(cuda::warp) / 2);
REQUIRE(result.unit_rank() == cuda::gpu_thread.rank(cuda::warp) / 2);
const auto lane_mask_ref =
(cuda::gpu_thread.rank(cuda::warp) % 2 == 0)
? cuda::device::lane_mask{0x5555'5555u}
: cuda::device::lane_mask(0xaaaa'aaaau);
REQUIRE(result.lane_mask() == lane_mask_ref);
REQUIRE(result.is_valid());
static_assert(Result::is_always_exhaustive());
static_assert(!Result::is_always_contiguous());
}
}
}
struct TestKernel
{
template <class Config>
__device__ void operator()(const Config& config)
{
test_binary_partition(config);
}
};
} // namespace
C2H_TEST("Binary partition mapping", "[group]")
{
const auto device = cuda::devices[0];
const cuda::stream stream{device};
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims<8, 4>());
cuda::launch(stream, config, TestKernel{});
}
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims(dim3{8, 4}));
cuda::launch(stream, config, TestKernel{});
}
stream.sync();
}

View File

@@ -1,196 +0,0 @@
//===----------------------------------------------------------------------===//
//
// Part of CUDA Experimental in CUDA C++ Core Libraries,
// under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES.
//
//===----------------------------------------------------------------------===//
#include <cuda/devices>
#include <cuda/hierarchy>
#include <cuda/launch>
#include <cuda/std/cstddef>
#include <cuda/std/numeric>
#include <cuda/std/tuple>
#include <cuda/std/type_traits>
#include <cuda/std/utility>
#include <cuda/stream>
#include <cuda/warp>
#include <cuda/experimental/group.cuh>
#include "group_testing.cuh"
namespace
{
template <class Mapping1, class Mapping2, class Config>
__device__ void test_composite_mapping(const Mapping1& mapping1, const Mapping2& mapping2, Config config)
{
using Mapping = cudax::composite_mapping<Mapping1, Mapping2>;
// Test construction from 2 mappings.
{
cudax::composite_mapping mapping{mapping1, mapping2};
static_assert(cuda::std::is_same_v<decltype(mapping), Mapping>);
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, Mapping1, Mapping2>
== (cuda::std::is_nothrow_copy_constructible_v<Mapping1>
&& cuda::std::is_nothrow_copy_constructible_v<Mapping2>) );
}
// Test get().
{
const cudax::composite_mapping mapping{mapping1, mapping2};
static_assert(cuda::std::is_same_v<decltype(mapping.get()), const cuda::std::tuple<Mapping1, Mapping2>&>);
static_assert(noexcept(mapping.get()));
const auto& mapping1_ref = cuda::std::get<0>(mapping.get());
CHECK(mapping1_ref.unit_count() == 4);
const auto& mapping2_ref = cuda::std::get<1>(mapping.get());
CHECK(mapping2_ref.unit_count(0) == 1);
CHECK(mapping2_ref.unit_count(1) == 3);
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
const cudax::composite_mapping mapping{mapping1, mapping2};
static_assert(
cudax::__group_mapping_result<decltype(mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result))>);
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
const auto rank_in_warp = cuda::gpu_thread.rank_as<unsigned>(parent_group);
if constexpr (Mapping1::static_unit_count() != cuda::std::dynamic_extent
&& Mapping2::static_group_count() != cuda::std::dynamic_extent)
{
static_assert(Result::static_group_count() == 16);
}
else
{
static_assert(Result::static_group_count() == cuda::std::dynamic_extent);
}
CHECK(result.group_count() == 16);
CHECK(result.group_rank() == (rank_in_warp / 4 * 2 + (rank_in_warp % 4 > 0)));
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
CHECK(result.unit_count() == ((rank_in_warp % 4 > 0) ? 3 : 1));
CHECK(result.unit_rank() == ((rank_in_warp % 4 > 0) ? (rank_in_warp % 4 - 1) : 0));
const auto lane_mask_ref = ((rank_in_warp % 4 > 0) ? 0b1110u : 0b0001u) << ((rank_in_warp / 4) * 4);
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
CHECK(result.is_valid());
static_assert(Result::is_always_exhaustive());
static_assert(Result::is_always_contiguous());
}
// Test operator|.
{
auto mapping = mapping1 | mapping2;
static_assert(cuda::std::is_same_v<Mapping, decltype(mapping)>);
static_assert(noexcept(mapping1 | mapping2));
const auto& mapping1_ref = cuda::std::get<0>(mapping.get());
CHECK(mapping1_ref.unit_count() == 4);
const auto& mapping2_ref = cuda::std::get<1>(mapping.get());
CHECK(mapping2_ref.unit_count(0) == 1);
CHECK(mapping2_ref.unit_count(1) == 3);
}
{
auto mapping = cudax::composite_mapping{mapping1} | mapping2;
static_assert(cuda::std::is_same_v<Mapping, decltype(mapping)>);
static_assert(noexcept(cudax::composite_mapping{mapping1} | mapping2));
const auto& mapping1_ref = cuda::std::get<0>(mapping.get());
CHECK(mapping1_ref.unit_count() == 4);
const auto& mapping2_ref = cuda::std::get<1>(mapping.get());
CHECK(mapping2_ref.unit_count(0) == 1);
CHECK(mapping2_ref.unit_count(1) == 3);
}
{
auto mapping = mapping1 | cudax::composite_mapping{mapping2};
static_assert(cuda::std::is_same_v<Mapping, decltype(mapping)>);
static_assert(noexcept(mapping1 | cudax::composite_mapping{mapping2}));
const auto& mapping1_ref = cuda::std::get<0>(mapping.get());
CHECK(mapping1_ref.unit_count() == 4);
const auto& mapping2_ref = cuda::std::get<1>(mapping.get());
CHECK(mapping2_ref.unit_count(0) == 1);
CHECK(mapping2_ref.unit_count(1) == 3);
}
{
auto mapping = cudax::composite_mapping{mapping1} | cudax::composite_mapping{mapping2};
static_assert(cuda::std::is_same_v<Mapping, decltype(mapping)>);
static_assert(noexcept(cudax::composite_mapping{mapping1} | cudax::composite_mapping{mapping2}));
const auto& mapping1_ref = cuda::std::get<0>(mapping.get());
CHECK(mapping1_ref.unit_count() == 4);
const auto& mapping2_ref = cuda::std::get<1>(mapping.get());
CHECK(mapping2_ref.unit_count(0) == 1);
CHECK(mapping2_ref.unit_count(1) == 3);
}
}
struct TestKernel
{
template <class Config>
__device__ void operator()(const Config& config)
{
{
const cudax::group_by<4> mapping1{};
const cudax::group_as mapping2{cuda::std::integer_sequence<cuda::std::size_t, 1, 3>{}};
test_composite_mapping(mapping1, mapping2, config);
}
{
const cudax::group_by mapping1{4};
const cudax::group_as mapping2{cuda::std::integer_sequence<cuda::std::size_t, 1, 3>{}};
test_composite_mapping(mapping1, mapping2, config);
}
{
const cudax::group_by<4> mapping1{};
constexpr unsigned counts2[]{1, 3};
const cudax::group_as mapping2{counts2};
test_composite_mapping(mapping1, mapping2, config);
}
{
const cudax::group_by mapping1{4};
constexpr unsigned counts2[]{1, 3};
const cudax::group_as mapping2{counts2};
test_composite_mapping(mapping1, mapping2, config);
}
}
};
} // namespace
C2H_TEST("Composite mapping", "[group]")
{
const auto device = cuda::devices[0];
const cuda::stream stream{device};
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims<8, 4>());
cuda::launch(stream, config, TestKernel{});
}
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims(dim3{8, 4}));
cuda::launch(stream, config, TestKernel{});
}
stream.sync();
}

View File

@@ -1,528 +0,0 @@
//===----------------------------------------------------------------------===//
//
// Part of CUDA Experimental in CUDA C++ Core Libraries,
// under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES.
//
//===----------------------------------------------------------------------===//
#include <cuda/devices>
#include <cuda/hierarchy>
#include <cuda/launch>
#include <cuda/std/cstddef>
#include <cuda/std/numeric>
#include <cuda/std/type_traits>
#include <cuda/std/utility>
#include <cuda/stream>
#include <cuda/warp>
#include <cuda/experimental/group.cuh>
#include "group_testing.cuh"
namespace
{
template <cuda::std::size_t... Ns, class Config>
__device__ void test_group_as(Config config)
{
using NsSeq = cuda::std::integer_sequence<cuda::std::size_t, Ns...>;
constexpr unsigned ns[]{static_cast<unsigned>(Ns)...};
constexpr cuda::std::size_t ngroups = sizeof...(Ns);
cuda::std::size_t group_starts[ngroups];
cuda::std::exclusive_scan(cuda::std::begin(ns), cuda::std::end(ns), group_starts, cuda::std::size_t{});
// Test static Ns.
{
using Mapping = cudax::group_as<cudax::__group_as_static_tag<Ns...>, true>;
// Test default constructor.
{
static_assert(cuda::std::is_trivially_default_constructible_v<Mapping>);
static_assert(cuda::std::is_empty_v<Mapping>);
Mapping mapping;
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(mapping.unit_count(i) == ns[i]);
}
}
// Test the mapping is constructible from the Ns sequence.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, NsSeq>);
cudax::group_as mapping{NsSeq{}};
static_assert(cuda::std::is_same_v<decltype(mapping), Mapping>);
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(mapping.unit_count(i) == ns[i]);
}
}
// Test the mapping is not constructible from Ns sequence and non_exhaustive_t.
static_assert(!cuda::std::is_constructible_v<Mapping, NsSeq, cudax::non_exhaustive_t>);
// Test static_group_count().
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_group_count())>);
static_assert(noexcept(Mapping::static_group_count()));
static_assert(Mapping::static_group_count() == ngroups);
// Test static_unit_count().
{
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_unit_count(cuda::std::size_t{}))>);
static_assert(noexcept(Mapping::static_unit_count(cuda::std::size_t{})));
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(Mapping::static_unit_count(i) == ns[i]);
}
}
// Test is_always_exhaustive().
static_assert(cuda::std::is_same_v<bool, decltype(Mapping::is_always_exhaustive())>);
static_assert(noexcept(Mapping::is_always_exhaustive()));
static_assert(Mapping::is_always_exhaustive());
// Test unit_count().
{
static_assert(
cuda::std::is_same_v<unsigned, decltype(cuda::std::declval<const Mapping>().unit_count(cuda::std::size_t{}))>);
static_assert(noexcept(cuda::std::declval<const Mapping>().unit_count(cuda::std::size_t{})));
const Mapping mapping;
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(mapping.unit_count(i) == ns[i]);
}
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping;
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
const auto rank_in_warp = cuda::gpu_thread.rank(parent_group);
unsigned group_rank_ref = ngroups - 1;
unsigned rank_ref = rank_in_warp - group_starts[group_rank_ref];
for (unsigned i = 1; i < ngroups; ++i)
{
if (rank_in_warp < group_starts[i])
{
group_rank_ref = i - 1;
rank_ref = rank_in_warp - group_starts[i - 1];
break;
}
}
static_assert(Result::static_group_count() == ngroups);
CHECK(result.group_count() == static_cast<unsigned>(ngroups));
CHECK(result.group_rank() == group_rank_ref);
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
CHECK(result.unit_count() == ns[group_rank_ref]);
CHECK(result.unit_rank() == rank_ref);
const auto lane_mask_ref =
(ns[group_rank_ref] < 32) ? ((1u << ns[group_rank_ref]) - 1) << group_starts[group_rank_ref] : ~0;
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
CHECK(result.is_valid());
static_assert(Result::is_always_exhaustive());
static_assert(Result::is_always_contiguous());
}
}
// Test dynamic Ns.
{
using Mapping = cudax::group_as<cudax::__group_as_dynamic_tag<ngroups>, true>;
// Test default constructor.
static_assert(!cuda::std::is_default_constructible_v<Mapping>);
// Test the mapping is constructible from the Ns array.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, decltype(ns)>);
cudax::group_as mapping{ns};
static_assert(cuda::std::is_same_v<decltype(mapping), Mapping>);
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(mapping.unit_count(i) == ns[i]);
}
}
// Test the mapping is not constructible from Ns array and non_exhaustive_t.
static_assert(!cuda::std::is_constructible_v<Mapping, decltype(ns), cudax::non_exhaustive_t>);
// Test static_group_count().
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_group_count())>);
static_assert(noexcept(Mapping::static_group_count()));
static_assert(Mapping::static_group_count() == ngroups);
// Test static_unit_count().
{
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_unit_count(cuda::std::size_t{}))>);
static_assert(noexcept(Mapping::static_unit_count(cuda::std::size_t{})));
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(Mapping::static_unit_count(i) == cuda::std::dynamic_extent);
}
}
// Test is_always_exhaustive().
static_assert(cuda::std::is_same_v<bool, decltype(Mapping::is_always_exhaustive())>);
static_assert(noexcept(Mapping::is_always_exhaustive()));
static_assert(Mapping::is_always_exhaustive());
// Test unit_count().
{
static_assert(
cuda::std::is_same_v<unsigned, decltype(cuda::std::declval<const Mapping>().unit_count(cuda::std::size_t{}))>);
static_assert(noexcept(cuda::std::declval<const Mapping>().unit_count(cuda::std::size_t{})));
const Mapping mapping{ns};
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(mapping.unit_count(i) == ns[i]);
}
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping{ns};
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
const auto rank_in_warp = cuda::gpu_thread.rank_as<unsigned>(parent_group);
unsigned group_rank_ref = ngroups - 1;
unsigned rank_ref = rank_in_warp - group_starts[group_rank_ref];
for (unsigned i = 1; i < ngroups; ++i)
{
if (rank_in_warp < group_starts[i])
{
group_rank_ref = i - 1;
rank_ref = rank_in_warp - group_starts[i - 1];
break;
}
}
static_assert(Result::static_group_count() == ngroups);
CHECK(result.group_count() == static_cast<unsigned>(ngroups));
CHECK(result.group_rank() == group_rank_ref);
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
CHECK(result.unit_count() == ns[group_rank_ref]);
CHECK(result.unit_rank() == rank_ref);
const auto lane_mask_ref =
(ns[group_rank_ref] < 32) ? ((1u << ns[group_rank_ref]) - 1) << group_starts[group_rank_ref] : ~0;
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
CHECK(result.is_valid());
static_assert(Result::is_always_exhaustive());
static_assert(Result::is_always_contiguous());
}
}
}
template <cuda::std::size_t... Ns, class Config>
__device__ void test_group_as_non_exhaustive(Config config)
{
using NsSeq = cuda::std::integer_sequence<cuda::std::size_t, Ns...>;
constexpr unsigned ns[]{static_cast<unsigned>(Ns)...};
constexpr cuda::std::size_t ngroups = sizeof...(Ns);
cuda::std::size_t group_starts[ngroups];
cuda::std::exclusive_scan(cuda::std::begin(ns), cuda::std::end(ns), group_starts, cuda::std::size_t{});
const auto ns_sum = cuda::std::accumulate(cuda::std::begin(ns), cuda::std::end(ns), 0u);
// Test static Ns.
{
using Mapping = cudax::group_as<cudax::__group_as_static_tag<Ns...>, false>;
// Test default constructor.
{
static_assert(cuda::std::is_trivially_default_constructible_v<Mapping>);
static_assert(cuda::std::is_empty_v<Mapping>);
Mapping mapping;
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(mapping.unit_count(i) == ns[i]);
}
}
// Test the mapping is not constructible from the Ns sequence.
static_assert(!cuda::std::is_constructible_v<Mapping, NsSeq>);
// Test the mapping is constructible from Ns sequence and non_exhaustive_t.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, NsSeq, cudax::non_exhaustive_t>);
cudax::group_as mapping{NsSeq{}, cudax::non_exhaustive};
static_assert(cuda::std::is_same_v<decltype(mapping), Mapping>);
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(mapping.unit_count(i) == ns[i]);
}
}
// Test static_group_count().
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_group_count())>);
static_assert(noexcept(Mapping::static_group_count()));
static_assert(Mapping::static_group_count() == ngroups);
// Test static_unit_count().
{
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_unit_count(cuda::std::size_t{}))>);
static_assert(noexcept(Mapping::static_unit_count(cuda::std::size_t{})));
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(Mapping::static_unit_count(i) == ns[i]);
}
}
// Test is_always_exhaustive().
static_assert(cuda::std::is_same_v<bool, decltype(Mapping::is_always_exhaustive())>);
static_assert(noexcept(Mapping::is_always_exhaustive()));
static_assert(!Mapping::is_always_exhaustive());
// Test unit_count().
{
static_assert(
cuda::std::is_same_v<unsigned, decltype(cuda::std::declval<const Mapping>().unit_count(cuda::std::size_t{}))>);
static_assert(noexcept(cuda::std::declval<const Mapping>().unit_count(cuda::std::size_t{})));
const Mapping mapping;
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(mapping.unit_count(i) == ns[i]);
}
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping;
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
const auto rank_in_warp = cuda::gpu_thread.rank(parent_group);
const auto is_valid_ref = (rank_in_warp < ns_sum);
static_assert(Result::static_group_count() == ngroups);
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
static_assert(!Result::is_always_exhaustive());
static_assert(Result::is_always_contiguous());
CHECK(result.group_count() == static_cast<unsigned>(ngroups));
CHECK(result.is_valid() == is_valid_ref);
if (is_valid_ref)
{
unsigned group_rank_ref = ngroups - 1;
unsigned rank_ref = rank_in_warp - group_starts[group_rank_ref];
for (unsigned i = 1; i < ngroups; ++i)
{
if (rank_in_warp < group_starts[i])
{
group_rank_ref = i - 1;
rank_ref = rank_in_warp - group_starts[i - 1];
break;
}
}
CHECK(result.group_rank() == group_rank_ref);
CHECK(result.unit_count() == ns[group_rank_ref]);
CHECK(result.unit_rank() == rank_ref);
const auto lane_mask_ref =
(ns[group_rank_ref] < 32) ? ((1u << ns[group_rank_ref]) - 1) << group_starts[group_rank_ref] : ~0;
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
}
}
}
// Test dynamic Ns.
{
using Mapping = cudax::group_as<cudax::__group_as_dynamic_tag<ngroups>, false>;
// Test default constructor.
static_assert(!cuda::std::is_default_constructible_v<Mapping>);
// Test the mapping is not constructible from the Ns array.
static_assert(!cuda::std::is_constructible_v<Mapping, decltype(ns)>);
// Test the mapping is constructible from Ns array and non_exhaustive_t.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, decltype(ns), cudax::non_exhaustive_t>);
cudax::group_as mapping{ns, cudax::non_exhaustive};
static_assert(cuda::std::is_same_v<decltype(mapping), Mapping>);
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(mapping.unit_count(i) == ns[i]);
}
}
// Test static_group_count().
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_group_count())>);
static_assert(noexcept(Mapping::static_group_count()));
static_assert(Mapping::static_group_count() == ngroups);
// Test static_unit_count().
{
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_unit_count(cuda::std::size_t{}))>);
static_assert(noexcept(Mapping::static_unit_count(cuda::std::size_t{})));
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(Mapping::static_unit_count(i) == cuda::std::dynamic_extent);
}
}
// Test is_always_exhaustive().
static_assert(cuda::std::is_same_v<bool, decltype(Mapping::is_always_exhaustive())>);
static_assert(noexcept(Mapping::is_always_exhaustive()));
static_assert(!Mapping::is_always_exhaustive());
// Test unit_count().
{
static_assert(
cuda::std::is_same_v<unsigned, decltype(cuda::std::declval<const Mapping>().unit_count(cuda::std::size_t{}))>);
static_assert(noexcept(cuda::std::declval<const Mapping>().unit_count(cuda::std::size_t{})));
const Mapping mapping{ns, cudax::non_exhaustive};
for (cuda::std::size_t i = 0; i < ngroups; ++i)
{
CHECK(mapping.unit_count(i) == ns[i]);
}
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping{ns, cudax::non_exhaustive};
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
const auto rank_in_warp = cuda::gpu_thread.rank(parent_group);
const auto is_valid_ref = (rank_in_warp < ns_sum);
static_assert(Result::static_group_count() == ngroups);
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
static_assert(!Result::is_always_exhaustive());
static_assert(Result::is_always_contiguous());
CHECK(result.group_count() == static_cast<unsigned>(ngroups));
CHECK(result.is_valid() == is_valid_ref);
if (is_valid_ref)
{
unsigned group_rank_ref = ngroups - 1;
unsigned rank_ref = rank_in_warp - group_starts[group_rank_ref];
for (unsigned i = 1; i < ngroups; ++i)
{
if (rank_in_warp < group_starts[i])
{
group_rank_ref = i - 1;
rank_ref = rank_in_warp - group_starts[i - 1];
break;
}
}
CHECK(result.group_rank() == group_rank_ref);
CHECK(result.unit_count() == ns[group_rank_ref]);
CHECK(result.unit_rank() == rank_ref);
const auto lane_mask_ref =
(ns[group_rank_ref] < 32) ? ((1u << ns[group_rank_ref]) - 1) << group_starts[group_rank_ref] : ~0;
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
}
}
}
}
struct TestKernel
{
template <class Config>
__device__ void operator()(const Config& config)
{
test_group_as<1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1>(
config);
test_group_as<2, 4, 8, 16, 2>(config);
test_group_as<3, 5, 1, 1, 22>(config);
test_group_as<31, 1>(config);
test_group_as<32>(config);
test_group_as_non_exhaustive<1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1>(
config);
test_group_as_non_exhaustive<2, 4, 8, 16, 2>(config);
test_group_as_non_exhaustive<3, 5, 1, 1, 22>(config);
test_group_as_non_exhaustive<31, 1>(config);
test_group_as_non_exhaustive<32>(config);
test_group_as_non_exhaustive<31>(config);
test_group_as_non_exhaustive<4, 6, 8>(config);
test_group_as_non_exhaustive<2, 2, 3, 1, 14>(config);
}
};
} // namespace
C2H_TEST("Group-as mapping", "[group]")
{
const auto device = cuda::devices[0];
const cuda::stream stream{device};
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims<8, 4>());
cuda::launch(stream, config, TestKernel{});
}
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims(dim3{8, 4}));
cuda::launch(stream, config, TestKernel{});
}
stream.sync();
}

View File

@@ -1,377 +0,0 @@
//===----------------------------------------------------------------------===//
//
// Part of CUDA Experimental in CUDA C++ Core Libraries,
// under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES.
//
//===----------------------------------------------------------------------===//
#include <cuda/devices>
#include <cuda/hierarchy>
#include <cuda/launch>
#include <cuda/std/cstddef>
#include <cuda/std/type_traits>
#include <cuda/std/utility>
#include <cuda/stream>
#include <cuda/warp>
#include <cuda/experimental/group.cuh>
#include "group_testing.cuh"
namespace
{
template <cuda::std::size_t N, class Config>
__device__ void test_group_by(Config config)
{
// Test static N.
{
using Mapping = cudax::group_by<N>;
static_assert(cuda::std::is_same_v<Mapping, cudax::group_by<N, true>>);
// Test default constructor.
{
static_assert(cuda::std::is_trivially_default_constructible_v<Mapping>);
static_assert(cuda::std::is_empty_v<Mapping>);
cudax::group_by<N> mapping;
CHECK(mapping.unit_count() == static_cast<unsigned>(N));
}
// Test the mapping is not constructible from unsigned.
static_assert(!cuda::std::is_constructible_v<Mapping, unsigned>);
// Test the mapping is not constructible from non_exhaustive_t.
static_assert(!cuda::std::is_constructible_v<Mapping, cudax::non_exhaustive_t>);
// Test the mapping is not constructible from unsigned and non_exhaustive_t.
static_assert(!cuda::std::is_constructible_v<Mapping, unsigned, cudax::non_exhaustive_t>);
// Test static_unit_count().
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_unit_count())>);
static_assert(noexcept(Mapping::static_unit_count()));
static_assert(Mapping::static_unit_count() == N);
// Test is_always_exhaustive().
static_assert(cuda::std::is_same_v<bool, decltype(Mapping::is_always_exhaustive())>);
static_assert(noexcept(Mapping::is_always_exhaustive()));
static_assert(Mapping::is_always_exhaustive());
// Test unit_count().
{
static_assert(cuda::std::is_same_v<unsigned, decltype(cuda::std::declval<const Mapping>().unit_count())>);
static_assert(noexcept(cuda::std::declval<const Mapping>().unit_count()));
const Mapping mapping;
CHECK(mapping.unit_count() == static_cast<unsigned>(N));
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping;
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
static_assert(Result::static_group_count() == 32 / N);
CHECK(result.group_count() == cuda::gpu_thread.count(cuda::warp) / N);
CHECK(result.group_rank() == cuda::gpu_thread.rank(cuda::warp) / N);
static_assert(Result::static_unit_count() == N);
CHECK(result.unit_count() == N);
CHECK(result.unit_rank() == cuda::gpu_thread.rank(cuda::warp) % N);
const auto lane_mask_ref = ((N < 32) ? ((1u << N) - 1) : ~0u) << ((cuda::gpu_thread.rank(cuda::warp) / N) * N);
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
CHECK(result.is_valid());
static_assert(Result::is_always_exhaustive());
static_assert(Result::is_always_contiguous());
}
}
// Test dynamic N.
{
using Mapping = cudax::group_by<>;
static_assert(cuda::std::is_same_v<Mapping, cudax::group_by<cuda::std::dynamic_extent, true>>);
// Test default constructor.
static_assert(!cuda::std::is_default_constructible_v<Mapping>);
static_assert(!cuda::std::is_empty_v<Mapping>);
// Test the mapping is constructible from unsigned.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, unsigned>);
cudax::group_by mapping{N};
static_assert(cuda::std::is_same_v<Mapping, decltype(mapping)>);
CHECK(mapping.unit_count() == static_cast<unsigned>(N));
}
// Test the mapping is not constructible from non_exhaustive_t.
static_assert(!cuda::std::is_constructible_v<Mapping, cudax::non_exhaustive_t>);
// Test the mapping is not constructible from unsigned and non_exhaustive_t.
static_assert(!cuda::std::is_constructible_v<Mapping, unsigned, cudax::non_exhaustive_t>);
// Test static_unit_count().
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_unit_count())>);
static_assert(noexcept(Mapping::static_unit_count()));
static_assert(Mapping::static_unit_count() == cuda::std::dynamic_extent);
// Test is_always_exhaustive().
static_assert(cuda::std::is_same_v<bool, decltype(Mapping::is_always_exhaustive())>);
static_assert(noexcept(Mapping::is_always_exhaustive()));
static_assert(Mapping::is_always_exhaustive());
// Test unit_count().
{
static_assert(cuda::std::is_same_v<unsigned, decltype(cuda::std::declval<const Mapping>().unit_count())>);
static_assert(noexcept(cuda::std::declval<const Mapping>().unit_count()));
const Mapping mapping{N};
CHECK(mapping.unit_count() == static_cast<unsigned>(N));
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping{N};
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
static_assert(Result::static_group_count() == cuda::std::dynamic_extent);
CHECK(result.group_count() == cuda::gpu_thread.count(cuda::warp) / N);
CHECK(result.group_rank() == cuda::gpu_thread.rank(cuda::warp) / N);
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
CHECK(result.unit_count() == N);
CHECK(result.unit_rank() == cuda::gpu_thread.rank(cuda::warp) % N);
const auto lane_mask_ref = ((N < 32) ? ((1u << N) - 1) : ~0u) << ((cuda::gpu_thread.rank(cuda::warp) / N) * N);
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
CHECK(result.is_valid());
static_assert(Result::is_always_exhaustive());
static_assert(Result::is_always_contiguous());
}
}
}
template <cuda::std::size_t N, class Config>
__device__ void test_group_by_non_exhaustive(Config config)
{
// Test static N.
{
using Mapping = cudax::group_by<N, false>;
// Test default constructor.
static_assert(cuda::std::is_trivially_default_constructible_v<Mapping>);
static_assert(cuda::std::is_empty_v<Mapping>);
// Test the mapping is not constructible from unsigned.
static_assert(!cuda::std::is_constructible_v<Mapping, unsigned>);
// Test the mapping is not constructible from non_exhaustive_t.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, cudax::non_exhaustive_t>);
Mapping mapping{cudax::non_exhaustive};
static_assert(cuda::std::is_same_v<decltype(mapping), Mapping>);
CHECK(mapping.unit_count() == static_cast<unsigned>(N));
}
// Test the mapping is not constructible from unsigned and non_exhaustive_t.
static_assert(!cuda::std::is_constructible_v<Mapping, unsigned, cudax::non_exhaustive_t>);
// Test static_unit_count().
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_unit_count())>);
static_assert(noexcept(Mapping::static_unit_count()));
static_assert(Mapping::static_unit_count() == N);
// Test is_always_exhaustive().
static_assert(cuda::std::is_same_v<bool, decltype(Mapping::is_always_exhaustive())>);
static_assert(noexcept(Mapping::is_always_exhaustive()));
static_assert(!Mapping::is_always_exhaustive());
// Test unit_count().
{
static_assert(cuda::std::is_same_v<unsigned, decltype(cuda::std::declval<const Mapping>().unit_count())>);
static_assert(noexcept(cuda::std::declval<const Mapping>().unit_count()));
const Mapping mapping{cudax::non_exhaustive};
CHECK(mapping.unit_count() == static_cast<unsigned>(N));
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping{cudax::non_exhaustive};
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
static_assert(Result::static_group_count() == 32 / N);
static_assert(Result::static_unit_count() == N);
static_assert(!Result::is_always_exhaustive());
static_assert(Result::is_always_contiguous());
const auto is_valid_ref = cuda::gpu_thread.rank(cuda::warp) < (cuda::gpu_thread.count(cuda::warp) / N) * N;
CHECK(result.is_valid() == is_valid_ref);
if (is_valid_ref)
{
CHECK(result.group_count() == cuda::gpu_thread.count(cuda::warp) / N);
CHECK(result.group_rank() == cuda::gpu_thread.rank(cuda::warp) / N);
CHECK(result.unit_count() == N);
CHECK(result.unit_rank() == cuda::gpu_thread.rank(cuda::warp) % N);
const auto lane_mask_ref = ((N < 32) ? ((1u << N) - 1) : ~0u) << ((cuda::gpu_thread.rank(cuda::warp) / N) * N);
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
}
}
}
// Test dynamic N.
{
using Mapping = cudax::group_by<cuda::std::dynamic_extent, false>;
// Test default constructor.
static_assert(!cuda::std::is_default_constructible_v<Mapping>);
static_assert(!cuda::std::is_empty_v<Mapping>);
// Test the mapping is constructible from unsigned.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, unsigned>);
Mapping mapping{N};
static_assert(cuda::std::is_same_v<Mapping, decltype(mapping)>);
CHECK(mapping.unit_count() == static_cast<unsigned>(N));
}
// Test the mapping is not constructible from non_exhaustive_t.
static_assert(!cuda::std::is_constructible_v<Mapping, cudax::non_exhaustive_t>);
// Test the mapping is not constructible from unsigned and non_exhaustive_t.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, unsigned, cudax::non_exhaustive_t>);
cudax::group_by mapping{static_cast<unsigned>(N), cudax::non_exhaustive};
static_assert(cuda::std::is_same_v<decltype(mapping), Mapping>);
CHECK(mapping.unit_count() == static_cast<unsigned>(N));
}
// Test static_unit_count().
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_unit_count())>);
static_assert(noexcept(Mapping::static_unit_count()));
static_assert(Mapping::static_unit_count() == cuda::std::dynamic_extent);
// Test unit_count().
{
static_assert(cuda::std::is_same_v<unsigned, decltype(cuda::std::declval<const Mapping>().unit_count())>);
static_assert(noexcept(cuda::std::declval<const Mapping>().unit_count()));
const Mapping mapping{N, cudax::non_exhaustive};
CHECK(mapping.unit_count() == static_cast<unsigned>(N));
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping{N, cudax::non_exhaustive};
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
static_assert(Result::static_group_count() == cuda::std::dynamic_extent);
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
static_assert(!Result::is_always_exhaustive());
static_assert(Result::is_always_contiguous());
const auto is_valid_ref = cuda::gpu_thread.rank(cuda::warp) < (cuda::gpu_thread.count(cuda::warp) / N) * N;
CHECK(result.is_valid() == is_valid_ref);
if (is_valid_ref)
{
CHECK(result.group_count() == cuda::gpu_thread.count(cuda::warp) / N);
CHECK(result.group_rank() == cuda::gpu_thread.rank(cuda::warp) / N);
CHECK(result.unit_count() == N);
CHECK(result.unit_rank() == cuda::gpu_thread.rank(cuda::warp) % N);
const auto lane_mask_ref = ((N < 32) ? ((1u << N) - 1) : ~0u) << ((cuda::gpu_thread.rank(cuda::warp) / N) * N);
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
}
}
}
}
struct TestKernel
{
template <class Config>
__device__ void operator()(const Config& config)
{
test_group_by<1>(config);
test_group_by<2>(config);
test_group_by<4>(config);
test_group_by<16>(config);
test_group_by<32>(config);
test_group_by_non_exhaustive<1>(config);
test_group_by_non_exhaustive<2>(config);
test_group_by_non_exhaustive<3>(config);
test_group_by_non_exhaustive<4>(config);
test_group_by_non_exhaustive<14>(config);
test_group_by_non_exhaustive<16>(config);
test_group_by_non_exhaustive<30>(config);
test_group_by_non_exhaustive<32>(config);
}
};
} // namespace
C2H_TEST("Group-by mapping", "[group]")
{
const auto device = cuda::devices[0];
const cuda::stream stream{device};
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims<8, 4>());
cuda::launch(stream, config, TestKernel{});
}
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims(dim3{8, 4}));
cuda::launch(stream, config, TestKernel{});
}
stream.sync();
}

View File

@@ -1,101 +0,0 @@
//===----------------------------------------------------------------------===//
//
// Part of CUDA Experimental in CUDA C++ Core Libraries,
// under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES.
//
//===----------------------------------------------------------------------===//
#include <cuda/devices>
#include <cuda/hierarchy>
#include <cuda/launch>
#include <cuda/std/cstddef>
#include <cuda/std/type_traits>
#include <cuda/std/utility>
#include <cuda/stream>
#include <cuda/warp>
#include <cuda/experimental/group.cuh>
#include "group_testing.cuh"
namespace
{
template <class Config>
__device__ void test_identity_mapping(Config config)
{
using Mapping = cudax::identity_mapping;
// Test default constructor.
{
static_assert(cuda::std::is_trivially_default_constructible_v<Mapping>);
static_assert(cuda::std::is_empty_v<Mapping>);
[[maybe_unused]] cudax::identity_mapping mapping;
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping;
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
static_assert(cuda::std::is_same_v<Result, ThreadsInWarpMappingResult>);
static_assert(Result::static_group_count() == ThreadsInWarpMappingResult::static_group_count());
CHECK(result.group_count() == prev_mapping_result.group_count());
CHECK(result.group_rank() == prev_mapping_result.group_rank());
static_assert(Result::static_unit_count() == ThreadsInWarpMappingResult::static_unit_count());
CHECK(result.unit_count() == prev_mapping_result.unit_count());
CHECK(result.unit_rank() == prev_mapping_result.unit_rank());
CHECK(result.lane_mask() == prev_mapping_result.lane_mask());
CHECK(result.is_valid() == prev_mapping_result.is_valid());
static_assert(Result::is_always_exhaustive() == ThreadsInWarpMappingResult::is_always_exhaustive());
static_assert(Result::is_always_contiguous() == ThreadsInWarpMappingResult::is_always_contiguous());
}
}
struct TestKernel
{
template <class Config>
__device__ void operator()(const Config& config)
{
test_identity_mapping(config);
test_identity_mapping(config);
test_identity_mapping(config);
test_identity_mapping(config);
test_identity_mapping(config);
}
};
} // namespace
C2H_TEST("Identity mapping", "[group]")
{
const auto device = cuda::devices[0];
const cuda::stream stream{device};
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims<8, 4>());
cuda::launch(stream, config, TestKernel{});
}
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims(dim3{8, 4}));
cuda::launch(stream, config, TestKernel{});
}
stream.sync();
}

View File

@@ -1,212 +0,0 @@
//===----------------------------------------------------------------------===//
//
// Part of CUDA Experimental in CUDA C++ Core Libraries,
// under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES.
//
//===----------------------------------------------------------------------===//
#include <cuda/devices>
#include <cuda/hierarchy>
#include <cuda/launch>
#include <cuda/std/cstddef>
#include <cuda/std/numeric>
#include <cuda/std/type_traits>
#include <cuda/std/utility>
#include <cuda/stream>
#include <cuda/warp>
#include <cuda/experimental/group.cuh>
#include "group_testing.cuh"
namespace
{
template <cuda::std::size_t N, class Config>
__device__ void test_take(Config config)
{
constexpr auto n = static_cast<unsigned>(N);
// Test static N.
{
using Mapping = cudax::take<N>;
// Test default constructor.
{
static_assert(cuda::std::is_trivially_default_constructible_v<Mapping>);
static_assert(cuda::std::is_empty_v<Mapping>);
Mapping mapping;
CHECK(mapping.unit_count() == n);
}
// Test static_unit_count().
{
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_unit_count())>);
static_assert(noexcept(Mapping::static_unit_count()));
static_assert(Mapping::static_unit_count() == N);
}
// Test unit_count().
{
static_assert(cuda::std::is_same_v<unsigned, decltype(cuda::std::declval<const Mapping>().unit_count())>);
static_assert(noexcept(cuda::std::declval<const Mapping>().unit_count()));
const Mapping mapping;
CHECK(mapping.unit_count() == n);
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping;
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
static_assert(Result::static_group_count() == ThreadsInWarpMappingResult::static_group_count());
static_assert(Result::static_unit_count() == N);
static_assert(Result::is_always_exhaustive()
== (Result::static_unit_count() == ThreadsInWarpMappingResult::static_unit_count()));
static_assert(Result::is_always_contiguous());
const auto is_valid_ref = cuda::std::cmp_less(cuda::gpu_thread.rank(cuda::warp), n);
CHECK(result.is_valid() == is_valid_ref);
if (is_valid_ref)
{
CHECK(result.group_count() == prev_mapping_result.group_count());
CHECK(result.group_rank() == prev_mapping_result.group_rank());
CHECK(result.unit_count() == n);
CHECK(result.unit_rank() == cuda::gpu_thread.rank(cuda::warp));
const auto lane_mask_ref = ((N < 32) ? ((1u << N) - 1) : ~0u);
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
CHECK(result.is_valid());
}
}
}
// Test dynamic Ns.
{
using Mapping = cudax::take<cuda::std::dynamic_extent>;
// Test default constructor.
{
static_assert(cuda::std::is_nothrow_default_constructible_v<Mapping>);
Mapping mapping;
CHECK(mapping.unit_count() == 0);
}
// Test the mapping is constructible from n.
{
static_assert(cuda::std::is_nothrow_constructible_v<Mapping, unsigned>);
cudax::take mapping{n};
static_assert(cuda::std::is_same_v<decltype(mapping), Mapping>);
CHECK(mapping.unit_count() == n);
}
// Test static_unit_count().
{
static_assert(cuda::std::is_same_v<cuda::std::size_t, decltype(Mapping::static_unit_count())>);
static_assert(noexcept(Mapping::static_unit_count()));
static_assert(Mapping::static_unit_count() == cuda::std::dynamic_extent);
}
// Test unit_count().
{
static_assert(cuda::std::is_same_v<unsigned, decltype(cuda::std::declval<const Mapping>().unit_count())>);
static_assert(noexcept(cuda::std::declval<const Mapping>().unit_count()));
const Mapping mapping{n};
CHECK(mapping.unit_count() == n);
}
// Test map(...).
{
const cudax::this_warp parent_group{config};
const ThreadsInWarpMappingResult prev_mapping_result;
static_assert(cudax::__group_mapping_result<decltype(cuda::std::declval<const Mapping>().map(
cuda::gpu_thread, parent_group, prev_mapping_result))>);
static_assert(
noexcept(cuda::std::declval<const Mapping>().map(cuda::gpu_thread, parent_group, prev_mapping_result)));
const Mapping mapping{n};
auto result = mapping.map(cuda::gpu_thread, parent_group, prev_mapping_result);
using Result = decltype(result);
static_assert(Result::static_group_count() == ThreadsInWarpMappingResult::static_group_count());
static_assert(Result::static_unit_count() == cuda::std::dynamic_extent);
static_assert(!Result::is_always_exhaustive());
static_assert(Result::is_always_contiguous());
const auto is_valid_ref = cuda::std::cmp_less(cuda::gpu_thread.rank(cuda::warp), n);
CHECK(result.is_valid() == is_valid_ref);
if (is_valid_ref)
{
CHECK(result.group_count() == prev_mapping_result.group_count());
CHECK(result.group_rank() == prev_mapping_result.group_rank());
CHECK(result.unit_count() == n);
CHECK(result.unit_rank() == cuda::gpu_thread.rank(cuda::warp));
const auto lane_mask_ref = ((N < 32) ? ((1u << N) - 1) : ~0u);
CHECK(result.lane_mask() == cuda::device::lane_mask{lane_mask_ref});
CHECK(result.is_valid());
}
}
}
}
struct TestKernel
{
template <class Config>
__device__ void operator()(const Config& config)
{
test_take<0>(config);
test_take<1>(config);
test_take<2>(config);
test_take<3>(config);
test_take<4>(config);
test_take<14>(config);
test_take<16>(config);
test_take<30>(config);
test_take<32>(config);
}
};
} // namespace
C2H_TEST("Take mapping", "[group]")
{
const auto device = cuda::devices[0];
const cuda::stream stream{device};
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims<8, 4>());
cuda::launch(stream, config, TestKernel{});
}
{
const auto config = cuda::make_config(cuda::grid_dims<1>(), cuda::block_dims(dim3{8, 4}));
cuda::launch(stream, config, TestKernel{});
}
stream.sync();
}